from airflow.models import BaseOperator
from pathlib import Path
import json
from bson import json_util
from aiflow.hooks.elastic_hook import ElasticHook
import logging
from jsonpath import jsonpath


logger = logging.getLogger(__name__)


class Elastic2CSVOperator(BaseOperator):
    """
    This operator allows you to read data from mongo, and write to a file
    """

    def __init__(self, elastic_conn_id, elastic_function, elastic_func_kwargs, output_fields, output_file, *args, **kwargs):
        super(Elastic2CSVOperator, self).__init__(*args, **kwargs)

        self.elastic_conn_id = elastic_conn_id
        self.elastic_function = elastic_function
        self.elastic_func_kwargs = elastic_func_kwargs
        self.output_fields = output_fields
        self.output_file = Path(output_file)

        self.elastic_search = kwargs.get('elastic_search', None)
        self.limit = kwargs.get('limit', None)

    def execute(self, context):
        logger.debug('start Elastic2CSVOperator ...')

        self.output_file.parent.mkdir(exist_ok=True, parents=True)
        if self.output_file.exists() and self.output_file.is_file:
            self.output_file.unlink()

        es = ElasticHook(self.elastic_conn_id).client()

        res = getattr(es, self.elastic_function)(**self.elastic_func_kwargs)
        hits = res['hits']
        if hits['total'] <= 0:
            logger.info(f'no hits found, no file will be exported. es result: {res}')
            return

        with open(self.output_file.absolute().as_posix(), 'w+') as f:
            f.write(','.join(self.output_fields) + '\n')
            for i, doc in enumerate(hits['hits']):
                if self.limit and i >= self.limit:
                    break

                # todo: escape the string in streaming writing way
                data = [self._value_of(doc['_source'], field, default='') for field in self.output_fields]
                data = list(map(lambda x: x if isinstance(x, str) else json.dumps(x), data))

                f.write(','.join(data) + '\n')
                f.flush()

                if i % 10000 == 0:
                    logger.debug(f'already wrote {i} lines data, limit: {self.limit}.')

    def _value_of(self, doc, field, default=None):
        try:
            found = jsonpath(doc, field)
            return found[0] if found else default
        except:
            return default



from airflow.models import BaseOperator
from pathlib import Path
import json
from bson import json_util
import logging
from collections import namedtuple
import pandas as pd
import re


logger = logging.getLogger(__name__)


class RegExLabellingOperator(BaseOperator):
    """
    This operator allows you to label data based on regex rules
    """

    class RegExRules:
        Rule = namedtuple('Ruled', ['regex', 'label'])

        def __init__(self, rules, unmatch_label, re_flags):
            self._unmatch_label = unmatch_label
            self._rules = [self.Rule(regex=r[0], label=r[1]) for r in rules]
            self._re_flags = re_flags

        def label(self, text):
            if not text or not isinstance(text, str):
                return self._unmatch_label

            return next((rule.label for rule in self._rules if re.search(rule.regex, text, self._re_flags)), self._unmatch_label)

    def __init__(self, input_file, data_column, regex_rules, label_column, output_file, *args, **kwargs):
        super(RegExLabellingOperator, self).__init__(*args, **kwargs)

        self.input_file = Path(input_file)
        self.data_column = data_column
        self.regex_rules = regex_rules
        self.label_column = label_column
        self.output_file = Path(output_file)

    def execute(self, context):
        logger.debug('start RegExLabellingOperator ...')

        assert self.input_file.name.lower().endswith('csv'), f'"csv" input file required'
        assert self.input_file.exists() and self.input_file.is_file(), f'invalid input_file: {input_file}'
        assert isinstance(self.regex_rules, self.RegExRules), f'regex_rules should be instance of RegExRules'

        self.output_file.parent.mkdir(exist_ok=True, parents=True)
        if self.output_file.exists() and self.output_file.is_file:
            self.output_file.unlink()

        df = pd.read_csv(self.input_file)
        df[self.label_column] = df[self.data_column].map(self.regex_rules.label)

        df.to_csv(self.output_file, index=False, header=True)



from airflow.models import BaseOperator
from pathlib import Path
import json
from bson import json_util
from aiflow.units.doc2mat_unit import Doc2MatUnit
import logging
import pandas as pd
import numpy as np
import shutil


logger = logging.getLogger(__name__)


class TextClassificationDataBuildOperator(BaseOperator):
    """
    This operator allows you to build dataset for text classification
    """

    def __init__(self, input_file, data_column, label_column, output_data_dir, output_extra_file, word2vec_file, *args, **kwargs):
        super(TextClassificationDataBuildOperator, self).__init__(*args, **kwargs)

        self.input_file = Path(input_file)
        self.data_column = data_column
        self.label_column = label_column
        self.output_data_dir = Path(output_data_dir)
        self.output_extra_file = Path(output_extra_file)

        n_max_words_count = kwargs.get('n_max_words_count', 5)
        self.doc2mat = Doc2MatUnit(n_max_words_count, word2vec_file, *args, **kwargs)

    def execute(self, context):
        logger.debug('start TextClassificationDataBuildOperator ...')

        assert self.input_file.name.lower().endswith('csv'), f'input file must be CSV.'
        assert self.input_file.exists() and self.input_file.is_file, f'invalid input_file: {input_file}'

        # prepare output env
        shutil.rmtree(self.output_data_dir, ignore_errors=True)
        self.output_data_dir.mkdir(exist_ok=True, parents=True)
        self.output_extra_file.parent.mkdir(exist_ok=True, parents=True)
        if self.output_extra_file.exists() and self.output_extra_file.is_file:
            self.output_extra_file.unlink()

        df = pd.read_csv(self.input_file)

        n_vector_dim = 300

        ohlabels = pd.get_dummies(df[self.label_column])
        with open(self.output_extra_file.absolute().as_posix(), 'w+') as f:
            f.write(json.dumps(dict(
                classes=list(ohlabels.columns)
            )))

        for i, row in df.iterrows():
            val = row[self.data_column]
            x = self.doc2mat.execute(sentence='' if pd.isna(val) else val)
            y = ohlabels.iloc[i].to_numpy()
            np.savez((self.output_data_dir / f'{i}.npz').absolute().as_posix(), x=x, y=y)
            if i % 10000 == 0:
                logger.debug(f'finish embedding X for {i} rows')

    def _prepare_output_dir(self, output_file):
        output_file.parent.mkdir(exist_ok=True, parents=True)
        if output_file.exists() and output_file.is_file:
            output_file.unlink()
        return self



import pkgutil
import logging
from abc import ABC, abstractmethod
import requests
import itertools
from multiprocessing import cpu_count
from concurrent.futures import ProcessPoolExecutor, as_completed


logger = logging.getLogger(__name__)


class Unit(ABC):

    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def execute(self, **kwargs):
        pass


__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module



"""
This file is forked from: https://github.com/airflow-plugins/mongo_plugin/blob/master/hooks/mongo_hook.py
"""

from ssl import CERT_NONE
from airflow.hooks.base_hook import BaseHook
from pymongo import MongoClient
import logging
from urllib.parse import quote_plus


logger = logging.getLogger(__name__)


class MongoHook(BaseHook):
    """
    PyMongo Wrapper to Interact With Mongo Database
    Mongo Connection Documentation
    https://docs.mongodb.com/manual/reference/connection-string/index.html
    You can specify connection string options in extra field of your connection
    https://docs.mongodb.com/manual/reference/connection-string/index.html#connection-string-options
    ex.
        {replicaSet: test, ssl: True, connectTimeoutMS: 30000}
    """
    conn_type = 'MongoDb'

    def __init__(self, conn_id='mongo_default', *args, **kwargs):
        super().__init__(source='mongo')

        self.mongo_conn_id = conn_id
        self.connection = self.get_connection(conn_id)
        self.extras = self.connection.extra_dejson

    def get_conn(self):
        """
        Fetches PyMongo Client
        """
        conn = self.connection

        uri = 'mongodb://{creds}{host}{port}/{database}'.format(
            creds='{}:{}@'.format(
                quote_plus(conn.login), quote_plus(conn.password)
            ) if conn.login is not None else '',

            host=conn.host,
            port='' if conn.port is None else ':{}'.format(conn.port),
            database='' if conn.schema is None else conn.schema
        )

        logger.debug(f'the mongodb connection uri is {uri}')

        # Mongo Connection Options dict that is unpacked when passed to MongoClient
        options = self.extras

        # If we are using SSL disable requiring certs from specific hostname
        if options.get('ssl', False):
            options.update({'ssl_cert_reqs': CERT_NONE})

        logger.debug(f'the mongodb connection options are {options}')
        return MongoClient(uri, **options)

    def get_collection(self, mongo_collection, mongo_db=None):
        """
        Fetches a mongo collection object for querying.
        Uses connection schema as DB unless specified
        """
        mongo_db = mongo_db if mongo_db is not None else self.connection.schema
        mongo_conn = self.get_conn()

        return mongo_conn.get_database(mongo_db).get_collection(mongo_collection)

    def aggregate(self, mongo_collection, aggregate_query, mongo_db=None, **kwargs):
        """
        Runs and aggregation pipeline and returns the results
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.aggregate
        http://api.mongodb.com/python/current/examples/aggregation.html
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.aggregate(aggregate_query, **kwargs)

    def find(self, mongo_collection, query, find_one=False, mongo_db=None, **kwargs):
        """
        Runs a mongo find query and returns the results
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.find
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        if find_one:
            return collection.find_one(query, **kwargs)
        else:
            return collection.find(query, **kwargs)

    def insert_one(self, mongo_collection, doc, mongo_db=None, **kwargs):
        """
        Inserts a single document into a mongo collection
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.insert_one
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.insert_one(doc, **kwargs)

    def insert_many(self, mongo_collection, docs, mongo_db=None, **kwargs):
        """
        Inserts many docs into a mongo collection.
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.insert_many
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.insert_many(docs, **kwargs)

    def replace_one(self, mongo_collection, replacement_filter, doc, mongo_db=None, **kwargs):
        """
        Replaces a single document that matches a filter in a mongo collection.
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.replace_one
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.replace_one(replacement_filter, doc, **kwargs)

    def update_one(self, mongo_collection, update_filter, update, mongo_db=None, **kwargs):
        """
        Updates a single document that matches a filter in a mongo collection.
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.update_one
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.update_one(update_filter, update, **kwargs)

    def update_many(self, mongo_collection, update_filter, update, mongo_db=None, **kwargs):
        """
        Updates many docs that matches a filter in a mongo collection.
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.update_many
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.update_many(update_filter, update, **kwargs)

    def bulk_write(self, mongo_collection, requests, mongo_db=None, **kwargs):
        """
        Submits a bulk write job mongo based on the specified requests.
        https://api.mongodb.com/python/current/api/pymongo/collection.html#pymongo.collection.Collection.bulk_write
        """
        collection = self.get_collection(mongo_collection, mongo_db=mongo_db)

        return collection.bulk_write(requests, **kwargs)



from ssl import CERT_NONE
from airflow.hooks.base_hook import BaseHook
import logging
from urllib.parse import quote_plus
from elasticsearch import Elasticsearch


logger = logging.getLogger(__name__)


class ElasticHook(BaseHook):

    conn_type = 'Elasticsearch'

    def __init__(self, conn_id='elastic_default', *args, **kwargs):
        super().__init__(source='elasticsearch')

        self.conn_id = conn_id
        self.connection = self.get_connection(conn_id)
        self.extras = self.connection.extra_dejson

    def client(self):

        uri = '{host}:{port}'.format(
            host=self.connection.host,
            port='' if self.connection.port is None else self.connection.port,
        )

        return Elasticsearch([uri])



from aiflow.units import Unit
import logging
import string
import re
from itertools import filterfalse
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from gensim.models import KeyedVectors
from pathlib import Path
import numpy as np


logger = logging.getLogger(__name__)


class Doc2VecUnit(Unit):

    def __init__(self, word2vec_model_url, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.lower = kwargs.get('lower', False)
        self.hyphen_replacement = kwargs.get('hyphen_replacement', None)
        self.remove_stopwords = kwargs.get('remove_stopwords', True)
        self.remove_numeric = kwargs.get('remove_numeric', True)
        self.lemmatize = kwargs.get('lemmatize', False)

        # todo: support load from website
        self.word2vec_file = word2vec_model_url
        self.word2vec = None

        self.n_ouput_dim = kwargs.get('n_ouput_dim', 300)

    def execute(self, **kwargs):

        # lazy load, otherwise it will time out when initialization
        if not self.word2vec:
            logger.debug('start loading word2vec model...')
            self.word2vec = KeyedVectors.load_word2vec_format(Path(self.word2vec_file).absolute().as_posix())
            logger.debug('finish loading word2vec model')

        words = self._sentence_to_words(kwargs.get('sentence', '').strip())
        vectors = list(filter(lambda x: x is not None, map(self._safe_word_to_vector, words)))
        if not vectors:
            return np.zeros(self.n_ouput_dim)

        return np.concatenate(vectors).reshape(len(vectors), -1).mean(axis=0)

    def _sentence_to_words(self, sentence):
        if not sentence or not sentence.strip():
            return sentence

        s = sentence.lower() if self.lower else sentence
        s = s.strip().replace('-', self.hyphen_replacement) if self.hyphen_replacement is not None else s

        words = re.split(r'[\s{}]'.format(string.punctuation), s)
        words = list(filterfalse(str.isnumeric, words)) if self.remove_numeric else words
        words = list(filter(lambda x: x not in stopwords.words('english'), words)) if self.remove_stopwords else words

        lemmatizer = WordNetLemmatizer()
        words = list(map(lemmatizer.lemmatize, words)) if self.lemmatize else words

        return list(filter(bool, words))

    def _safe_word_to_vector(self, word):
        if word in self.word2vec:
            return self.word2vec[word]
        else:
            return None
